@@ -1,6 +1,9 @@
+import itertools
 import logging
 import re
 
+import pandas as pd
+
 from .gtf import GTF
 from .utils import open_as_text
 
@@ -15,18 +18,24 @@ class FASTA:
     """
     PARSER = re.compile(r'^>(?P<sequence_id>\S+)(?P<group>.*)')
     GROUP_PARSER = re.compile(r'(?P<key>\S+?):(?P<value>\S+)')
-    BASEPAIRS = {
-        'a': 'T',
+    COMPLEMENT = {
         'A': 'T',
-        'c': 'G',
         'C': 'G',
-        'g': 'C',
         'G': 'C',
-        't': 'A',
         'T': 'A',
-        'n': 'N',
+        'Y': 'R',
+        'R': 'Y',
+        'W': 'W',
+        'S': 'S',
+        'K': 'M',
+        'M': 'K',
+        'D': 'H',
+        'V': 'B',
+        'H': 'D',
+        'B': 'V',
         'N': 'N',
     }
+    SEQUENCE_PARSER = re.compile(r'[^atcgATCG]')
 
     def __init__(self, fasta_path):
         self.fasta_path = fasta_path
@@ -78,11 +87,15 @@ class FASTA:
         :return: reverse complement
         :rtype: str
         """
-        return ''.join(FASTA.BASEPAIRS[b] for b in reversed(sequence))
+        return ''.join(FASTA.COMPLEMENT[b] for b in reversed(sequence.upper()))
 
-    def entries(self):
+    def entries(self, parse=True):
         """Generator that yields one FASTA entry (sequence ID + sequence) at a time.
 
+        :param parse: whether or not to parse the header into a dictionary,
+                      defaults to `True`
+        :type parse: bool, optional
+
         :return: a generator that yields a tuple of the FASTA entry
         :rtype: generator
         """
@@ -95,7 +108,7 @@ class FASTA:
                         yield info, sequence
                         sequence = ''
 
-                    info = FASTA.parse_header(line)
+                    info = FASTA.parse_header(line) if parse else line.strip()
                 else:
                     sequence += line.strip()
 
@@ -107,7 +120,11 @@ class FASTA:
 
         :param out_path: path to generate the sorted FASTA
         :type out_path: str
+
+        :return: path to sorted FASTA file, set of chromosomes in FASTA file
+        :rtype: tuple
         """
+        chromosomes = set()
         to_sort = []
         with open_as_text(self.fasta_path, 'r') as f:
             position = 0
@@ -115,9 +132,12 @@ class FASTA:
 
             while line:
                 if line.startswith('>'):
-                    to_sort.append([
-                        FASTA.parse_header(line)['sequence_id'], position, None
-                    ])
+                    header = FASTA.parse_header(line)
+                    logger.debug(
+                        'Found FASTA entry {}'.format(header['sequence_id'])
+                    )
+                    to_sort.append([header['sequence_id'], position, None])
+                    chromosomes.add(header['sequence_id'])
 
                 position = f.tell()
                 line = f.readline()
@@ -131,6 +151,7 @@ class FASTA:
         with open_as_text(self.fasta_path,
                           'r') as fasta, open_as_text(out_path, 'w') as f:
             for tup in to_sort:
+                logger.debug(f'Writing FASTA entry {tup[0]}')
                 start_position = tup[1]
                 end_position = tup[2]
                 fasta.seek(start_position)
@@ -141,8 +162,120 @@ class FASTA:
                     for line in fasta:
                         f.write(line)
 
+        return out_path, chromosomes
 
-def generate_cdna_fasta(fasta_path, gtf_path, out_path):
+
+def generate_kite_fasta(feature_path, out_path, no_mismatches=False):
+    """Generate a FASTA file for feature barcoding with the KITE workflow.
+
+    This FASTA contains all sequences that are 1 hamming distance from the
+    provided barcodes. The file of barcodes must be a 2-column TSV containing
+    the barcode sequences in the first column and their corresponding feature
+    name in the second column. If hamming distance 1 variants collide for any
+    pair of barcodes, the hamming distance 1 variants for those barcodes are
+    not generated.
+
+    :param feature_path: path to TSV containing barcodes and feature names
+    :type feature_path: str
+    :param out_path: path to FASTA to generate
+    :type out_path: str
+    :param no_mismatches: whether to generate hamming distance 1 variants,
+                          defaults to `False`
+    :type no_mismatches: bool, optional
+
+    :raises Exception: if there are barcodes of different lengths
+    :raises Exception: if there are duplicate barcodes
+
+    :return: (path to generated FASTA, set of barcode lengths)
+    :rtype: tuple
+    """
+
+    def generate_mismatches(name, sequence):
+        """Helper function to generate 1 hamming distance mismatches.
+
+        :param name: name of the sequence
+        :type name: str
+        :param sequence: the sequence
+        :type sequence: str
+
+        :return: a generator that yields a tuple of (unique name, mismatched sequence)
+        :rtype: generator
+        """
+        sequence = sequence.upper()
+        for i in range(len(sequence)):
+            base = sequence[i]
+            before = sequence[:i]
+            after = sequence[i + 1:]
+
+            for j, different in enumerate([b for b in ['A', 'C', 'G', 'T']
+                                           if b != base]):
+                yield f'{name}-{i}.{j+1}', f'{before}{different}{after}'
+
+    df_features = pd.read_csv(
+        feature_path, sep='\t', header=None, names=['sequence', 'name']
+    )
+
+    lengths = set()
+    features = {}
+    variants = {}
+    # Generate all feature barcode variations before saving to check for collisions.
+    for i, row in df_features.iterrows():
+        # Check that the first column contains the sequence
+        # and the second column the feature name.
+        if FASTA.SEQUENCE_PARSER.search(row.sequence.upper()):
+            raise Exception((
+                'Encountered non-ATCG basepairs in barcode sequence {row.sequence}. '
+                'Does the first column contain the sequences and the second column the feature names?'
+            ))
+
+        lengths.add(len(row.sequence))
+        features[row['name']] = row.sequence
+        variants[row['name']] = {
+            name: seq
+            for name, seq in generate_mismatches(row['name'], row.sequence)
+            if not no_mismatches
+        }
+
+    # Check duplicate barcodes.
+    duplicates = set([
+        bc for bc in features.values() if list(features.values()).count(bc) > 1
+    ])
+    if len(duplicates) > 0:
+        raise Exception(
+            'Duplicate feature barcodes: {}'.format(' '.join(duplicates))
+        )
+    if len(lengths) > 1:
+        logger.warning(
+            'Detected barcodes of different lengths: {}'.format(
+                ','.join(str(l) for l in lengths)  # noqa
+            )
+        )
+    for f1, f2 in itertools.combinations(variants.keys(), 2):
+        v1 = variants[f1]
+        v2 = variants[f2]
+        if len(set.intersection(set(v1.values()), set(v2.values()))) > 0:
+            logger.warning((
+                f'Detected features with barcodes that are not >2 hamming distance apart: {f1}, {f2}. '
+                f'Hamming distance 1 variants for these barcodes will not be generated.'
+            ))
+            variants[f1] = {}
+            variants[f2] = {}
+
+    # Write FASTA
+    with open_as_text(out_path, 'w') as f:
+        for feature, barcode in features.items():
+            attributes = [('feature_id', feature)]
+            f.write(f'{FASTA.make_header(feature, attributes)}\n')
+            f.write(f'{barcode}\n')
+
+            for name, variant in variants[feature].items():
+                f.write(f'{FASTA.make_header(name, attributes)}\n')
+                f.write(f'{variant}\n')
+
+    return out_path, min(lengths)
+
+
+def generate_cdna_fasta(fasta_path, gtf_path, out_path, chromosomes=None):
     """Generate a cDNA FASTA using the genome and GTF.
 
     This function assumes the order in which the chromosomes appear in the
@@ -155,6 +288,10 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
     :type gtf_path: str
     :param out_path: path to cDNA FASTA to generate
     :type out_path: str
+    :param chromosomes: set of chromosomes to generate sequences for. If not
+                        provided, sequences for all chromosomes are generated
+                        by default, defaults to `None`
+    :type chromosomes: set, optional
 
     :return: path to generated cDNA FASTA
     :rtype: str
@@ -163,10 +300,14 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
     gtf = GTF(gtf_path)
     gtf_entries = gtf.entries()
 
+    count = 0
     with open_as_text(out_path, 'w') as f:
         previous_gtf_entry = None
         for info, sequence in fasta.entries():
             sequence_id = info['sequence_id']
+            if chromosomes and sequence_id not in chromosomes:
+                continue
+
             logger.debug(
                 'Generating cDNA from chromosome {}'.format(sequence_id)
             )
@@ -175,13 +316,13 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
             transcript_infos = {}
             while True:
                 try:
-                    gtf_entry = previous_gtf_entry if previous_gtf_entry else next(
-                        gtf_entries
-                    )
+                    gtf_entry = previous_gtf_entry or next(gtf_entries)
                 except StopIteration:
                     break
                 previous_gtf_entry = None
                 chromosome = gtf_entry['seqname']
+                if chromosomes and chromosome not in chromosomes:
+                    continue
 
                 if sequence_id != chromosome:
                     previous_gtf_entry = gtf_entry
@@ -217,11 +358,15 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
                         gene_id, gene_version
                     ) if gene_version else gene_id
                     gene_name = gtf_entry['group'].get('gene_name', '')
+                    transcript_name = gtf_entry['group'].get(
+                        'transcript_name', ''
+                    )
 
                     if transcript not in transcript_infos:
                         attributes = [
                             ('gene_id', gene),
                             ('gene_name', gene_name),
+                            ('transcript_name', transcript_name),
                             ('chr', chromosome),
                             ('start', start),
                             ('end', end),
@@ -230,8 +375,28 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
                         transcript_infos[transcript] = attributes
 
             logger.debug(
-                'Writing {} cDNA transcripts'.format(len(transcript_sequences))
+                f'Writing {len(transcript_sequences)} cDNA transcripts for chromosome {chromosome}'
             )
+            # Validation.
+            info_unique = set(transcript_infos.keys()
+                              ) - set(transcript_sequences.keys())
+            sequences_unique = set(transcript_sequences.keys()
+                                   ) - set(transcript_infos.keys())
+            if info_unique:
+                raise Exception(
+                    'The following transcripts have "transcript" feature(s) but no '
+                    'corresponding "exon" feature: {}'.format(
+                        ', '.join(info_unique)
+                    )
+                )
+            if sequences_unique:
+                raise Exception(
+                    'The following transcripts have a "exon" feature but no '
+                    'corresponding "transcript" feature(s): {}'.format(
+                        ', '.join(sequences_unique)
+                    )
+                )
+
             for transcript in sorted(transcript_sequences.keys()):
                 exon = transcript_sequences[transcript]
                 attributes = transcript_infos[transcript]
@@ -244,11 +409,15 @@ def generate_cdna_fasta(fasta_path, gtf_path, out_path):
                         '+' else FASTA.reverse_complement(exon)
                     )
                 )
+                count += 1
+    logger.info(f'Wrote {count} cDNA transcripts')
 
     return out_path
 
 
-def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
+def generate_intron_fasta(
+    fasta_path, gtf_path, out_path, chromosomes=None, flank=30
+):
     """Generate an intron FASTA using the genome and GTF.
 
     This function assumes the order in which the chromosomes appear in the
@@ -267,6 +436,10 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
     :type gtf_path: str
     :param out_path: path to intron FASTA to generate
     :type out_path: str
+    :param chromosomes: set of chromosomes to generate sequences for. If not
+                        provided, sequences for all chromosomes are generated
+                        by default, defaults to `None`
+    :type chromosomes: set, optional
     :param flank: the size of intron flanks, in bases, defaults to `30`
     :type flank: int, optional
 
@@ -277,10 +450,13 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
     gtf = GTF(gtf_path)
     gtf_entries = gtf.entries()
 
+    count = 0
     with open_as_text(out_path, 'w') as f:
         previous_gtf_entry = None
         for info, sequence in fasta.entries():
             sequence_id = info['sequence_id']
+            if chromosomes and sequence_id not in chromosomes:
+                continue
             logger.debug(
                 'Generating introns from chromosome {}'.format(sequence_id)
             )
@@ -289,13 +465,13 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
             transcript_infos = {}
             while True:
                 try:
-                    gtf_entry = previous_gtf_entry if previous_gtf_entry else next(
-                        gtf_entries
-                    )
+                    gtf_entry = previous_gtf_entry or next(gtf_entries)
                 except StopIteration:
                     break
                 previous_gtf_entry = None
                 chromosome = gtf_entry['seqname']
+                if chromosomes and chromosome not in chromosomes:
+                    continue
 
                 if sequence_id != chromosome:
                     previous_gtf_entry = gtf_entry
@@ -330,11 +506,15 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
                         gene_id, gene_version
                     ) if gene_version else gene_id
                     gene_name = gtf_entry['group'].get('gene_name', '')
+                    transcript_name = gtf_entry['group'].get(
+                        'transcript_name', ''
+                    )
 
                     if transcript not in transcript_infos:
                         attributes = [
                             ('gene_id', gene),
                             ('gene_name', gene_name),
+                            ('transcript_name', transcript_name),
                             ('chr', chromosome),
                             ('start', start),
                             ('end', end),
@@ -342,6 +522,29 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
                         ]
                         transcript_infos[transcript] = attributes
 
+            logger.debug((
+                f'Found {len(transcript_exons)} transcripts with exons for chromosome {chromosome}.'
+                'Generating introns from these transcripts'
+            ))
+            # Validation.
+            info_unique = set(transcript_infos.keys()
+                              ) - set(transcript_exons.keys())
+            sequences_unique = set(transcript_exons.keys()
+                                   ) - set(transcript_infos.keys())
+            if info_unique:
+                raise Exception(
+                    'The following transcripts have "exon" feature(s) but no '
+                    'corresponding "transcript" feature: {}'.format(
+                        ', '.join(info_unique)
+                    )
+                )
+            if sequences_unique:
+                raise Exception(
+                    'The following transcripts have a "transcript" feature but no '
+                    'corresponding "exon" feature(s): {}'.format(
+                        ', '.join(sequences_unique)
+                    )
+                )
             for transcript in sorted(transcript_exons.keys()):
                 attributes = transcript_infos[transcript]
 
@@ -394,6 +597,7 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
                                 '+' else FASTA.reverse_complement(intron)
                             )
                         )
+                        count += 1
                         index += 1
                         flank_start = max(start - flank, transcript_interval[0])
                         flank_end = min(end + flank, transcript_interval[1])
@@ -412,7 +616,9 @@ def generate_intron_fasta(fasta_path, gtf_path, out_path, flank=30):
                             '+' else FASTA.reverse_complement(intron)
                         )
                     )
+                    count += 1
                     index += 1
+    logger.info(f'Wrote {count} intron sequences')
 
     return out_path
 
